///\file main.cpp


//you'd really think they'd have fixed these compiler errors before releasing msvc7....
#pragma warning(disable: 4267) 
#pragma warning(disable: 4244) 

#include "main.h"
#include <sven/string_addons.h>
#include <sven/regex_addons.h>
#include <sven/array_types.h>
#include <sven/fmt.h>
#include <list>
#include <sven/foreach.h>

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/lu.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/fwd.hpp>
#include <boost/format.hpp>
#include <sstream>
#include <fstream>

typedef double Scalar;
typedef boost::numeric::ublas::vector<Scalar> Vector;
typedef boost::numeric::ublas::matrix<Scalar> Matrix;
typedef const Vector & cVector;
typedef const Matrix & cMatrix;

using boost::numeric::ublas::inner_prod;
using boost::numeric::ublas::norm_2;
using boost::numeric::ublas::trans;
//using boost::numeric::ublas::add;
using boost::numeric::ublas::prod;

//a non-memory safe function for initilizing a matrix from scalar memory.
void init(Matrix &m, Scalar * d) {
	for(int i=0;i<m.size1();i++) for(int j=0;j<m.size2();j++) 
		m(i,j)=d[i*m.size2()+j];
}

//a non-memory safe function for initilizing a vector from scalar memory.
void init(Vector &v, Scalar * d) {
	for(int i=0;i<v.size();i++) v[i]=d[i];
}

///boilerplate analytic solver.
void my_solve( Matrix A, Vector &x, cVector b, size_t n) {

	using namespace boost::numeric::ublas;
	permutation_matrix<double> P(n);
	lu_factorize(A,P);

	x = b;
	lu_substitute(A,P,x);
}

using namespace std;
using namespace sven;
using namespace sven::regex_addons;

struct KalmanFilter {

	list<double2> observations;
	list<double2> truth;
	typedef list<double2>::const_iterator lci_obs;
	typedef list<double2>::iterator li_obs;

	const Scalar dt;
	Matrix H, F, FT, I;
	Matrix Q, R;
	Matrix K;
	Vector w, v;
	list<Matrix> P;
	lci_obs current_z;
	list<Vector> S;
	
	//initilizations for the linear 2d motion case.
	KalmanFilter(const Scalar p_error,const Scalar r_error,const Scalar q_error) :
		//known values
		H(4,4), F(2,4), FT(4,2), Q(4,4), R(2,2), dt(1), I(4,4),
		
		//calculated values
		K(2,2) {
		
		{
			Scalar h[] = {
				1, 0, dt, 0,
				0, 1, 0, dt,
				0, 0, 1, 0,
				0, 0, 0, 1
			};
			
			init(H,h);
		}

		//model error
		{
		
			const Scalar er=q_error; 
			Scalar q[] = {
				er, 0, 0, 0,
				0, er, 0, 0,
				0, 0, er, 0,
				0, 0, 0, er
			};
			init(Q,q);
		}

		//measurment error
		{
			const Scalar er=r_error; 
			Scalar r[] = {
				er, 0,
				0, er 
			};
			
			init(R,r);
		}

		{
			Scalar i[] = {
				1, 0, 0, 0,
				0, 1, 0, 0,
				0, 0, 1, 0,
				0, 0, 0, 1
			};
			
			init(I,i);
		}
		
		//measurment operator
		{
			Scalar f[] = {
				1, 0, 0, 0,
				0, 1, 0, 0,
			};
			
			init(F,f);
			FT=trans(F);
		}
		
		//initial state.
		{
			Vector S1(4);
			Scalar s[] = {
				20, 20, 10, 10,
			};
			
			init(S1,s);
			S.push_back(S1);
		}
		{
			Matrix P1(4,4);
			const Scalar er=p_error; 
			Scalar p[] = {
				er, 0, 0, 0,
				0, er, 0, 0,
				0, 0, er, 0,
				0, 0, 0, er
			};
			init(P1,p);
			P.push_back(P1);
		}
		
	}

	void take_step() {
	
		#define debug_out(x) cout << #x << ": " << x << endl;
	
		double2 zd=*(current_z);
		current_z++;
		Vector z(2);
		z[0]=zd[0]; z[1]=zd[1];
		
		Vector S_last = S.back();
		
		
		//debug_out(S_last);
		Vector S_pre = prod(H,S_last);
		//debug_out(S_pre);
		Matrix P_last = P.back();
		//debug_out(P_last);
		//P_pre = HP_lastH^T + Q
		//debug_out(Q);
		Matrix P_pre = prod(Matrix(prod(H,P_last)), trans(H))  + Q;

	
		//debug_out(S_pre); 
		//debug_out(P_pre);
		
		//yi = zi - FS_pre
		Vector y(2);
		y = z - prod(F,S_pre);
		
		//debug_out(z);
		//debug_out(y);
		//debug_out(H);
		
		//Ui = FiPi|i-1FTi + R
		Matrix U(2,2);
		//debug_out(FT);
		Matrix temp=prod(P_pre,FT);
		//debug_out(temp);
		U=prod(F,temp) + R;
		//debug_out(U); 
		//debug_out(R);
		
		Scalar a=U(0,0), b=U(0,1), c=U(1,0), d=U(1,1);
		Matrix U_inv(2,2);
		U_inv(0,0)=d; U_inv(0,1)=-b;
		U_inv(1,0)=-c; U_inv(1,1)=a;
		Scalar denom=a*d-b*c;
		if(denom<.0001) RT_ERROR("singular U matrix");
		U_inv = (1.0/denom)*U_inv;
		
		//debug_out(U_inv);
		
		//Ki = Pi|i-1FTiU-1
		Matrix K(4,2);
		K=prod(P_pre,Matrix(prod(FT,U_inv)));
		
		//debug_out(K);
		
		Vector S_new=S_pre+prod(K,y);
		Matrix temp2=prod(K,F);
		temp2=I-temp2;
		//debug_out(y);
		Matrix P_new=prod(Matrix(temp2),P_pre);
		
		P.push_back(P_new);
		S.push_back(S_new);
		
		debug_out(S_new);		
		
	}

	void gen_truth() {
		double2 p(20,20);
		double2 v(10,10);

		for(int i=0;i<30;i++) {
			truth.push_back(p);
			p+=v;
		}
	}

	void load_observations() {
		observations.clear();
		ifstream in("observations.txt");
		while(in.good()) {
			double2 obs;
			in >> obs.x >> obs.y;
			observations.push_back(obs);
		}
	}

	void write_filter(cstring xfname, cstring yfname) {
		ofstream outx(xfname.c_str());
		ofstream outy(yfname.c_str());
		
		std::list<Vector>::iterator i;
		int c=1;
		foreach(i,S) {
			outx << c << " " << (*i)[0] << endl;
			outy << c << " " << (*i)[1] << endl;	
			c++;
		}
		
	}

	void write_data(const list<double2> &data, cstring xfname, cstring yfname) {
		ofstream outx(xfname.c_str());
		ofstream outy(yfname.c_str());
		
		lci_obs i;
		int c=1;
		foreach(i,data) {
			outx << c << " " << i->x << endl;
			outy << c << " " << i->y << endl;
			c++;
		}
	}


	string yobservation_file() { return (fmt("%s.observations.y.d") % name).str(); }
	string ytruth_file() { return (fmt("%s.truth.y.d") % name).str(); }
	string yfilter_file() { return (fmt("%s.filter.y.d") % name).str(); }
	string ymp_file() { return (fmt("%s.y.mp") % name).str();  }

	string xobservation_file() { return (fmt("%s.observations.x.d") % name).str(); }
	string xtruth_file() { return (fmt("%s.truth.x.d") % name).str(); }
	string xfilter_file() { return (fmt("%s.filter.x.d") % name).str(); }
	string xmp_file() { return (fmt("%s.x.mp") % name).str();  }

	string name;

	void gen_files() {
		xgen_mp();
		ygen_mp();
		write_data(observations,xobservation_file(),yobservation_file());
		write_data(truth,xtruth_file(),ytruth_file());	
		write_filter(xfilter_file(),yfilter_file());	
	}

	void xgen_mp() {
		
		write_file(xmp_file(),
		fmt(
		"input graph																\n"
		"beginfig(1);																\n"
		"draw begingraph(5.5in,2in);												\n"
		"gdraw %s withcolor blue;													\n"
		"gdraw %s withcolor black withpen pencircle scaled 1.2pt;					\n"
		"gdraw %s withcolor red withpen pencircle scaled 1pt;						\n"
		"glabel.bot(btex $x$-positions etex, OUT);									\n"
		"endgraph;																	\n"
		"endfig;																	\n"
		"end																		\n"
		) % qw(xobservation_file()) % qw(xtruth_file()) % qw(xfilter_file()));
	}

	void ygen_mp() {
		
		write_file(ymp_file(),
		fmt(
		"input graph																\n"
		"beginfig(1);																\n"
		"draw begingraph(5.5in,2in);												\n"
		"gdraw %s withcolor blue;													\n"
		"gdraw %s withcolor black withpen pencircle scaled 1.2pt;					\n"
		"gdraw %s withcolor red withpen pencircle scaled 1pt;						\n"
		"glabel.bot(btex $x$-positions etex, OUT);									\n"
		"endgraph;																	\n"
		"endfig;																	\n"
		"end																		\n"
		) % qw(yobservation_file()) % qw(ytruth_file()) % qw(yfilter_file()));
	}	
	

};

int system(cfmt f) { return system(f.str().c_str()); }

//initial error
//measurement error
//process error
void do_test(cstring name, Scalar p_error, Scalar r_error, Scalar q_error) {

	KalmanFilter f(p_error,r_error,q_error);

	f.load_observations();
	f.gen_truth();
	
	f.current_z=f.observations.begin();
	
	for(int i=0;i<29;i++) f.take_step();
	
	f.name=name;	
	f.gen_files();
	system(fmt("mp %s.x.mp") %name);
	system(fmt("mp %s.y.mp") %name);	
}

void main() {
try {

	SetCurrentDirectory("doc");

	do_test("t1",0.1,30,0);
	do_test("t2",10,10,0);
	do_test("t3",10,10,10);
	
}
catch(exception &e) {
	cout << "exception caught!" << endl << e.what() << endl;
}

}

